import os
import logging
import hashlib
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryFile
from shutil import copyfileobj
from functools import lru_cache

from .files import File
from . import exceptions

EXTRAS = []

try:
    import boto3
    from botocore.exceptions import ClientError
    EXTRAS.append(boto3)
except ImportError:
    pass

try:
    from google.cloud import storage
    from crc32c import crc32
    import base64
    EXTRAS.append(storage)
except ImportError:
    pass

if not EXTRAS:
    raise ImportError('At least one extra from [aws, gcp] must be installed.')

logger = logging.getLogger(__name__)


def strip_prefix(text, prefix):
    return text[len(prefix):] if text.startswith(prefix) else text


class BaseStorage:
    def __init__(self, file_class=None):
        if file_class and not issubclass(file_class, File):
            raise TypeError(f'FileCls must inherit from File')
        self.file_class = file_class or File
        self.hash_method = None

    def read_into_stream(self, path, stream=None):
        raise NotImplementedError()

    def open(self, path, mode='rb', encoding=None):
        logger.debug('Reading from %s', path)
        return self.file_class(path, storage=self, mode=mode, encoding=encoding)

    def create_bucket(self):
        raise NotImplementedError()

    def write(self, f, path):
        raise NotImplementedError()

    def list(self, path):
        raise NotImplementedError()


@lru_cache()
def get_s3_bucket(_storage, bucket):
    return _storage.s3.Bucket(bucket)


def s3etag(file_like,
           multipart_threshold=8 * 1024 * 1024,
           multipart_chunksize=8 * 1024 * 1024):
    """
    Returns the md5 hash that will match the ETag in S3
    https://stackoverflow.com/questions/6591047/etag-definition-changed-in-amazon-s3/28877788#28877788
    https://github.com/boto/boto3/blob/0cc6042615fd44c6822bd5be5a4019d0901e5dd2/boto3/s3/transfer.py#L169
    """
    md5hash = hashlib.md5()
    file_like.seek(0)
    filesize = 0
    block_count = 0
    md5string = b''
    for block in iter(lambda: file_like.read(multipart_chunksize), b''):
        md5hash = hashlib.md5()
        md5hash.update(block)
        md5string += md5hash.digest()
        filesize += len(block)
        block_count += 1

    if filesize > multipart_threshold:
        md5hash = hashlib.md5()
        md5hash.update(md5string)
        md5hash = md5hash.hexdigest() + "-" + str(block_count)
    else:
        md5hash = md5hash.hexdigest()

    file_like.seek(0)
    # https://github.com/aws/aws-sdk-net/issues/815
    return f'"{md5hash}"'


class S3Storage(BaseStorage):
    def __init__(self,
                 aws_access_key_id=None,
                 aws_secret_access_key=None,
                 aws_session_token=None,
                 region_name=None,
                 use_ssl=None,
                 endpoint_url=None,
                 bucket=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.bucket = bucket
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.region_name = region_name
        self.use_ssl = use_ssl
        self.endpoint_url = endpoint_url
        self._resource = None
        self.hash_method = s3etag

    @property
    def s3(self):
        logger.debug('Getting S3 resource')
        # See how boto resolve credentials in
        # http://boto3.readthedocs.io/en/latest/guide/configuration.html#guide-configuration
        if not self._resource:
            logger.debug('Resource does not exist, creating a new one...')
            self._resource = boto3.resource(
                's3',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                region_name=self.region_name,
                use_ssl=self.use_ssl,
                endpoint_url=self.endpoint_url,
            )
        return self._resource

    def read_into_stream(self, path, stream=None):
        try:
            stream = stream or BytesIO()
            bucket = get_s3_bucket(self, self.bucket)
            bucket.download_fileobj(path, stream)
            stream.seek(0)
            return stream
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                logger.debug('File %s in bucket %s does not exist', path, bucket)
                raise exceptions.FileDoesNotExist('s3://{}/{}'.format(self.bucket, path))
            else:
                raise

    def create_bucket(self):
        self.s3.create_bucket(Bucket=self.bucket)

    def write(self, f, path):
        logger.debug('Writing to %s', path)
        bucket = get_s3_bucket(self, self.bucket)
        # boto3 closes the handler, creating a copy
        # https://github.com/boto/s3transfer/issues/80
        with TemporaryFile() as file_to_close:
            f.seek(0)
            copyfileobj(f, file_to_close)
            file_to_close.seek(0)
            bucket.upload_fileobj(file_to_close, path)

    def list(self, path):
        logger.debug('Listing %s', path)
        bucket = get_s3_bucket(self, self.bucket)
        path = path.rstrip('/')
        try:
            for obj in bucket.objects.filter(Prefix=path):
                if not obj.key.endswith('/'):
                    yield obj.e_tag, obj.key
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchBucket':
                logger.warning('Bucket does not exist, list() returning empty.')
            else:
                raise


@lru_cache()
def get_gcs_bucket(_storage, bucket):
    return _storage.gcs.get_bucket(bucket)


def gcs_crc32c(file_like):
    """
    https://stackoverflow.com/questions/52686848/does-google-cloud-storage-client-in-python-check-crc-or-md5-automatically
    """
    # Read data and checksum - read whole file in memory right now
    # TODO: read in chuncks
    file_like.seek(0)
    checksum = crc32(file_like.read())
    # Convert into a bytes type that can be base64 encoded
    base64_crc32c = base64.b64encode(checksum.to_bytes(length=4, byteorder='big')).decode('utf-8')
    # Print the Base64 encoded CRC32C
    return base64_crc32c


class GCStorage(BaseStorage):
    def __init__(self, credential_file=None, bucket=None, **kwargs):
        super().__init__(**kwargs)
        self.bucket = bucket
        self.credential_file = credential_file
        self._resource = None
        self.hash_method = gcs_crc32c

    @property
    def gcs(self):
        logger.debug('Getting GCS resource')
        if not self._resource:
            logger.debug('Resource does not exist, creating a new one...')
            if self.credential_file:
                self._resource = storage.Client.from_service_account_json(self.credential_file)
            else:
                self._resource = storage.Client()
        return self._resource

    def read_into_stream(self, path, stream=None):
        stream = stream or BytesIO()
        bucket = get_gcs_bucket(self, self.bucket)
        blob = bucket.blob(path)
        blob.download_to_file(stream)
        stream.seek(0)
        return stream

    def create_bucket(self):
        self.gcs.create_bucket(self.bucket)

    def write(self, f, path):
        logger.debug('Writing to %s', path)
        bucket = get_gcs_bucket(self, self.bucket)
        blob = bucket.blob(path)
        f.seek(0)
        blob.upload_from_file(f, path)

    def list(self, path):
        logger.debug('Listing %s', path)
        bucket = get_gcs_bucket(self, self.bucket)
        path = path.rstrip('/')
        for obj in bucket.list_blobs(prefix=path):
            if not obj.name.endswith('/'):
                yield obj.crc32c, obj.name


class LocalStorage(BaseStorage):
    def __init__(self, root='/', hash_method=None, **kwargs):
        super().__init__(**kwargs)
        self.root = root
        self.hash_method = hash_method

    def build_path(self, path):
        path = path.lstrip('/')
        return os.path.join(self.root, path)

    def read_into_stream(self, path, stream=None):
        try:
            stream = stream or BytesIO()
            with open(path, 'rb') as f:
                stream.write(f.read())
            stream.seek(0)
            return stream
        except FileNotFoundError:
            raise exceptions.FileDoesNotExist(path)

    def create_bucket(self):
        pass

    def write(self, f, path):
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        f.seek(0)
        copyfileobj(f, open(path, 'wb'))

    def list(self, path):
        path = Path(path)
        if path.is_dir():
            for root, dirs, files in os.walk(path):
                for file in files:
                    file_hash = self.hash_method(open(Path(root).joinpath(file), 'rb')) if self.hash_method else None
                    yield file_hash, str(Path(root).joinpath(file))
        else:
            # only yields if it exists
            if path.exists():
                yield self.hash_method(open(path, 'rb')) if self.hash_method else None, str(path)



import logging
import subprocess
import shlex
import os
from pathlib import Path

import click
import click_log
from click.exceptions import UsageError, MissingParameter

from . import s3conf, config, __version__
from .exceptions import EnvfilePathNotDefinedError, EnvfilePathNotDefinedUsageError
from .storage.exceptions import FileDoesNotExist

logger = logging.getLogger(__name__)


class SectionArgument(click.Argument):
    def handle_parse_result(self, *args, **kwargs):
        try:
            return super().handle_parse_result(*args, **kwargs)
        except MissingParameter:
            raise EnvfilePathNotDefinedUsageError()


@click.group(invoke_without_command=True)
@click.version_option(version=__version__)
@click.option('--edit', '-e', is_flag=True)
@click.pass_context
# this sets the log level for this app only
@click_log.simple_verbosity_option('s3conf')
def main(ctx, edit):
    """
    Simple command line tool to help manage environment variables stored in a S3-like system. Facilitates editing text
    files remotely stored, as well as downloading and uploading files.
    """
    # configs this module logger to behave properly
    # logger messages will go to stderr (check __init__.py/patch.py)
    # client output should be generated with click.echo() to go to stdout
    try:
        click_log.basic_config('s3conf')
        logger.debug('Running main entrypoint')
        if edit:
            if ctx.invoked_subcommand is None:
                settings = config.Settings()
                conf = s3conf.S3Conf(settings=settings)
                logger.debug('Using config file %s', settings.config_file)
                storage = conf.storages.storage(settings.config_file)
                with storage.open(settings.config_file, 'r+') as config_file:
                    config_file.edit()
                return
            else:
                raise UsageError('Edit should not be called with a subcommand.')
        # manually call help in case no relevant settings were defined
        if ctx.invoked_subcommand is None:
            click.echo(main.get_help(ctx))
    except FileDoesNotExist as e:
        raise UsageError('The file {} does not exist. Try the "init" command if you want to create it.'.format(str(e)))


@main.command('env')
@click.argument('section',
                required=False)
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file.')
@click.option('--phusion',
              is_flag=True,
              help='If set, dumps variables to --dump-path in for format used by the phusion docker image. '
                   'More information in https://github.com/phusion/baseimage-docker.')
@click.option('--phusion-path',
              default='/etc/container_environment',
              show_default=True,
              help='Path where to dump variables as in the phusion docker image format.')
@click.option('--quiet',
              '-q',
              is_flag=True,
              help='Do not print any environment variables output.')
@click.option('--edit',
              '-e',
              is_flag=True)
@click.option('--create',
              '-c',
              is_flag=True,
              help='Create the remote environment file if it does not exist.')
def env(section, map_files, phusion, phusion_path, quiet, edit, create):
    """
    Reads the file defined by the S3CONF variable and output its contents to stdout. Logs are printed to stderr.
    See options for added functionality: editing file, mapping files, dumping in the phusion-baseimage format, etc.
    """
    try:
        logger.debug('Running env command')
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(settings=settings)

        if create or edit:
            if create:
                conf.create_envfile()
            if edit:
                conf.edit_envfile()
        else:
            with conf.get_envfile() as env_file:
                env_vars = env_file.as_dict()
            if map_files:
                conf.pull()
            if not quiet:
                for var_name, var_value in sorted(env_vars.items(), key=lambda x: x[0]):
                    click.echo('{}={}'.format(var_name, var_value))
            if phusion:
                s3conf.phusion_dump(env_vars, phusion_path)
    except EnvfilePathNotDefinedError:
        raise EnvfilePathNotDefinedUsageError()
    except FileDoesNotExist as e:
        raise UsageError('The file {} does not exist. Try "-c" option if you want to create it.'.format(str(e)))


@main.command('add')
@click.argument('section', cls=SectionArgument)
@click.argument('local_path')
def add(section, local_path):
    """
    Add a mapping to the S3CONF_MAP variable for the given SECTION pointing to LOCAL_PATH.
    The remote file is mapped to the folder "files" in the same folder of the file pointed by S3CONF.
    """
    try:
        settings = config.Settings(section=section)
        local_path = Path(local_path).resolve().relative_to(settings.root_folder)
        remote_path = os.path.join(os.path.dirname(settings.environment_file_path), 'files', str(local_path))
        settings.add_mapping(remote_path, str(local_path))
        config_file = config.ConfigFileResolver(settings.config_file, section=section)
        config_file.set('S3CONF_MAP', settings.serialize_mappings())
        config_file.save()
    except EnvfilePathNotDefinedError:
        raise EnvfilePathNotDefinedUsageError()


@main.command('rm')
@click.argument('section', cls=SectionArgument)
@click.argument('local_path')
def rm(section, local_path):
    """
    Removes the mapping in the S3CONF_MAP variable for the given SECTION pointing to LOCAL_PATH.
    """
    try:
        settings = config.Settings(section=section)
        local_path = Path(local_path).resolve().relative_to(settings.root_folder)
        settings.rm_mapping(str(local_path))
        config_file = config.ConfigFileResolver(settings.config_file, section=section)
        config_file.set('S3CONF_MAP', settings.serialize_mappings())
        config_file.save()
    except EnvfilePathNotDefinedError:
        raise EnvfilePathNotDefinedUsageError()


@main.command('push')
@click.argument('section', cls=SectionArgument)
@click.option('--force',
              '-f',
              is_flag=True)
def push(section, force):
    """
    Upload files mapped in S3CONF_MAP variable defined in s3conf.ini for the given section.
    Stores the md5 hash for uploaded files in local cache. If the remote file md5 hash differs
    from the value we have in our cache, the upload fails unless forced.
    """
    try:
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(settings=settings)
        conf.push(force=force)
    except EnvfilePathNotDefinedError:
        raise EnvfilePathNotDefinedUsageError()


@main.command('exec')
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file.')
@click.argument('section',
                required=False)
# to use option-like arguments use "--"
# http://click.pocoo.org/6/arguments/#option-like-arguments
@click.argument('command',
                required=False,
                nargs=-1,
                type=click.UNPROCESSED)
@click.pass_context
def exec_command(ctx, section, command, map_files):
    """
    Sets the process environemnt and executes the [COMMAND] in the same context. Does not modify the current shell
    environment.

    If the [COMMAND] has option-like arguments, use the standard POSIX pattern "--" to separate options
    from arguments. Considering our configuration in the "dev" section, we could write:

    s3conf -v info exec dev -- ping -v google.com
    """
    try:
        logger.debug('Running exec command')
        command = ' '.join(command)

        if not command:
            logger.warning('No command detected.')
            click.echo(exec_command.get_help(ctx))
            return

        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(settings=settings)
        with conf.get_envfile() as env_file:
            env_vars = env_file.as_dict()
        if map_files:
            conf.pull()

        current_env = os.environ.copy()
        current_env.update(env_vars)
        logger.debug('Executing command "%s"', command)
        subprocess.run(shlex.split(command), env=current_env, check=True)
    except EnvfilePathNotDefinedError:
        raise EnvfilePathNotDefinedUsageError()


@main.command('set')
@click.argument('section', cls=SectionArgument)
@click.argument('value',
                required=False)
def set_variable(section, value):
    """
    Set value of a variable in an environment file for the given section.
    If the variable is already defined, its value is replaced, otherwise, it is added to the end of the file.
    The value is given as "ENV_VAR_NAME=env_var_value", e.g.:

    s3conf set test ENV_VAR_NAME=env_var_value
    """
    if not value:
        value = section
        section = None
    try:
        logger.debug('Running env command')
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(settings=settings)

        with conf.get_envfile(mode='r+') as env_vars:
            env_vars.set(value)
    except EnvfilePathNotDefinedError:
        raise EnvfilePathNotDefinedUsageError()


@main.command('unset')
@click.argument('section', cls=SectionArgument)
@click.argument('value',
                required=False)
def unset_variable(section, value):
    """
    Unset a variable in an environment file for the given section.
    The value is given is the variable name, e.g.:

    s3conf unset test ENV_VAR_NAME
    """
    if not value:
        value = section
        section = None
    try:
        logger.debug('Running env command')
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(settings=settings)

        with conf.get_envfile(mode='r+') as env_vars:
            env_vars.unset(value)
    except EnvfilePathNotDefinedError:
        raise EnvfilePathNotDefinedUsageError()


@main.command('init')
@click.argument('section')
@click.argument('remote-file')
def init(section, remote_file):
    """
    Creates the .s3conf config folder and .s3conf/config config file
    with the provided section name and configuration file. It is a very
    basic config file. Manually edit it in order to add credentials. E.g.:

    s3conf init development s3://my-project/development.env
    """
    logger.debug('Running init command')
    settings = config.Settings(section=section)
    config_file = config.ConfigFileResolver(settings.config_file, section=section)
    config_file.set('S3CONF', remote_file)
    config_file.save()
    conf = s3conf.S3Conf(settings=settings)
    conf.verify_cache()
